@@ -21,9 +21,11 @@ import urllib
 # inhouse libraries
 import encode_utils.transfer_to_gcp
 import encode_utils as eu
-import encode_utils.profiles as eup
+from encode_utils.profiles import Profiles
 import encode_utils.utils as euu
 
+# EU-21 add support for attachment in autosql file type
+mimetypes.add_type('text/autosql', '.as')
 
 #: The directory that contains the log files created by the `Connection` class.
 LOG_DIR = "EU_Logs"
@@ -124,7 +126,7 @@ class Connection():
     #: Constant
     PATCH = "patch"
 
-    def __init__(self, dcc_mode=None, dry_run=False, submission=False):
+    def __init__(self, dcc_mode=None, dry_run=False, submission=False, no_log_file=False):
 
         #: A reference to the `debug` logging instance that was created earlier in ``encode_utils.debug_logger``.
         #: This class adds a file handler, such that all messages sent to it are logged to this
@@ -141,6 +143,7 @@ class Connection():
         self.dcc_mode = self._set_dcc_mode(dcc_mode)
         self.dcc_host = eu.DCC_MODES[self.dcc_mode]["host"]
         self.dcc_url = eu.DCC_MODES[self.dcc_mode]["url"]
+        self.profiles = Profiles(self.dcc_url)
 
         #: Set to True to prevent any server-side changes on the ENCODE Portal, i.e. PUT, POST,
         #: PATCH, DELETE requests will not be sent to the Portal. After-POST and after-PATCH
@@ -149,14 +152,10 @@ class Connection():
         #: :meth:`set_live_run`.
         self.dry_run = dry_run
 
-        # Add debug file handler to debug_logger:
-        self._add_file_handler(logger=self.debug_logger, level=logging.DEBUG, tag="debug")
-
         #: A ``logging`` instance with a file handler for logging terse error messages.
         #: The log file resides locally within the directory specified by the constant
         #: ``connection.LOG_DIR``. Accepts messages >= ``logging.ERROR``.
         self.error_logger = logging.getLogger(eu.ERROR_LOGGER_NAME)
-        self._add_file_handler(logger=self.error_logger, level=logging.ERROR, tag="error")
         self.log_error("Connecting to {}".format(self.dcc_host))
 
         #: A ``logging`` instance with a file handler for logging successful POST operations.
@@ -165,7 +164,12 @@ class Connection():
         self.post_logger = logging.getLogger(eu.POST_LOGGER_NAME)
         log_level = logging.INFO
         self.post_logger.setLevel(log_level)
-        self._add_file_handler(logger=self.post_logger, level=log_level, tag="posted")
+
+        # Add file handlers
+        if not no_log_file:
+            self._add_file_handler(logger=self.debug_logger, level=logging.DEBUG, tag="debug")
+            self._add_file_handler(logger=self.error_logger, level=logging.ERROR, tag="error")
+            self._add_file_handler(logger=self.post_logger, level=log_level, tag="posted")
 
         self.check_dry_run() #If on, signal this in the logs.
 
@@ -555,8 +559,8 @@ class Connection():
                 raise ProfileNotSpecified(
                     ("You need to specify the ID of the profile to submit to by using the '{}' key"
                      " in the payload, or by setting the `@id` property explicitely.").format(self.PROFILE_KEY))
-        profile = eup.Profile(profile_id)  # raises euu.UnknownProfile if unknown profile ID.
-        return profile.profile_id
+        profile = self.profiles.get_profile_from_id(profile_id)
+        return profile
 
     def get_lookup_ids_from_payload(self, payload):
         """
@@ -715,11 +719,11 @@ class Connection():
             rec_id: `str`. An identifier for the new File object on the Portal.
             profile_id: `str`. The ID of the profile that the record belongs to.
         """
-        if profile_id != eup.Profile.FILE_PROFILE_ID:
+        if profile_id != self.profiles.FILE_PROFILE_ID:
             return
         self.upload_file(file_id=rec_id)
 
-    def after_submit_hooks(self, rec_id, profile_id, method=""):
+    def after_submit_hooks(self, rec_id, profile_id, method="", upload_file=True):
         """
         Calls after-POST and after-PATCH hooks. This method is called from both the ``post()`` and
         ``patch()`` instance methods. Returns the None object immediately if the dry-run feature
@@ -734,6 +738,8 @@ class Connection():
             profile_id: `str`. The profile identifier indicating the profile that the record belongs to.
             method: str. One of ``self.POST`` or ``self.PATCH``, or the empty string to indicate which
                 registered hooks to look through.
+            upload_file: `bool`. If `False`, skip uploading files to the Portal.
+                Defaults to `True`.
         """
         if self.check_dry_run():
             return
@@ -749,7 +755,12 @@ class Connection():
         #... None yet.
 
         # Call POST-specific hooks if POST:
-        if method == self.POST:
+        if upload_file is False:
+            self.debug_logger.debug(
+                "Will not upload file %s to the portal since upload_file is False",
+                rec_id,
+            )
+        if method == self.POST and upload_file is True:
             self.after_submit_file_cloud_upload(rec_id, profile_id)
 
         # Call PATCH-specific hooks if PATCH:
@@ -774,7 +785,7 @@ class Connection():
             return payload
         aliases = euu.clean_aliases(payload[eu.ALIAS_PROP_NAME])
         aliases = self.add_alias_prefix(aliases)
-        payload[eu.ALIAS_PROP_NAME] = aliases
+        payload[eu.ALIAS_PROP_NAME] = sorted(set(aliases))
         return payload
 
     def before_submit_attachment(self, payload):
@@ -825,24 +836,24 @@ class Connection():
               `encode_utils.utils.calculate_md5sum` when it can't calculate the md5sum.
         """
         profile_id = payload[self.PROFILE_KEY]
-        if profile_id != eup.Profile.FILE_PROFILE_ID:
+        if profile_id != self.profiles.FILE_PROFILE_ID:
             return payload
         try:
-            file_name = payload[eup.Profile.SUBMITTED_FILE_PROP_NAME]
+            file_name = payload[self.profiles.SUBMITTED_FILE_PROP_NAME]
         except KeyError:
             return payload
         # Set md5sum
-        if (eup.Profile.MD5SUM_NAME_PROP_NAME in payload) and (payload[eup.Profile.MD5SUM_NAME_PROP_NAME]):
+        if (self.profiles.MD5SUM_NAME_PROP_NAME in payload) and (payload[self.profiles.MD5SUM_NAME_PROP_NAME]):
             # Already set; nothing to do.
             pass
         else:
-            payload[eup.Profile.MD5SUM_NAME_PROP_NAME] = euu.calculate_md5sum(file_name)
+            payload[self.profiles.MD5SUM_NAME_PROP_NAME] = euu.calculate_md5sum(file_name)
         # Set file_size
-        if (eup.Profile.FILE_SIZE_PROP_NAME in payload) and (payload[eup.Profile.FILE_SIZE_PROP_NAME]):
+        if (self.profiles.FILE_SIZE_PROP_NAME in payload) and (payload[self.profiles.FILE_SIZE_PROP_NAME]):
             # Already set; nothing to do.
             pass
         else:
-            payload[eup.Profile.FILE_SIZE_PROP_NAME] = euu.calculate_file_size(file_name)
+            payload[self.profiles.FILE_SIZE_PROP_NAME] = euu.calculate_file_size(file_name)
         return payload
 
     def before_post_fastq_file(self, payload):
@@ -858,7 +869,7 @@ class Connection():
 
         """
         profile_id = payload[self.PROFILE_KEY]
-        if profile_id != eup.Profile.FILE_PROFILE_ID:
+        if profile_id != self.profiles.FILE_PROFILE_ID:
             return payload
 
         run_type = payload.get("run_type")
@@ -911,7 +922,14 @@ class Connection():
 
         return payload
 
-    def post(self, payload, require_aliases=True):
+    def post(
+        self,
+        payload,
+        require_aliases=True,
+        upload_file=True,
+        return_original_status_code=False,
+        truncate_long_strings_in_payload_log=False,
+    ):
         """POST a record to the Portal.
 
         Requires that you include in the payload the non-schematic key ``self.PROFILE_KEY`` to
@@ -935,11 +953,24 @@ class Connection():
                  as many times as you want on the Portal when not providing an alias.  Furthermore,
                  submitting labs should include at least one alias per record being submitted
                  to the Portal for traceabilty purposes in the submitting lab.
+            upload_file: `bool`. If `False`, when POSTing files the file data will not
+                be uploaded to S3, defaults to `True`. This can be useful if you have
+                custom upload logic. If the files to upload are already on disk, it is
+                recommmended to leave this with the default, which will use `aws s3 cp`
+                to upload them.
+            return_original_status_code: `bool`. Defaults to `False`. If `True`, then
+                will return the original `requests.Response.status_code` of the initial
+                post, in addition to the usual `dict` response.
+            truncate_long_strings_in_payload_log: `bool`. Defaults to `False`. If
+                `True`, then long strings (> 1000 characters) present in the payload
+                will be truncated before being logged.
 
         Returns:
             `dict`: The JSON response from the POST operation, or the existing record if it already
             exists on the Portal (where a GET on any of it's aliases, when provided in the payload,
-            finds the existing record).
+            finds the existing record). If `return_original_status_code=True`, then will
+            return a `tuple` of the above `dict` and an `int` corresponding to the
+            status code on POST of the initial payload.
 
         Raises:
             encode_utils.connection.AwardPropertyMissing: The `award` property isn't present in the payload and there isn't a
@@ -958,15 +989,15 @@ class Connection():
         # Make sure we have a payload that can be converted to valid JSON, and
         # tuples become arrays, ...
         payload = json.loads(json.dumps(payload))
-        profile_id = self.get_profile_from_payload(payload)
-        payload[self.PROFILE_KEY] = profile_id
-        url = euu.url_join([self.dcc_url, profile_id])
+        profile = self.get_profile_from_payload(payload)
+        payload[self.PROFILE_KEY] = profile.name
+        url = euu.url_join([self.dcc_url, profile.name])
         if self.ENCID_KEY in payload:
             # Shouldn't be here, unless maybe a PATCH was attempted and the record didn't exist, so
             # a POST was then attempted.
             payload.pop(self.ENCID_KEY)
         # Check if we need to add defaults for 'award' and 'lab' properties:
-        if profile_id not in eup.Profile.AWARDLESS_PROFILE_IDS:  # No lab prop for these profiles either.
+        if profile.has_award:  # No lab prop for these profiles either.
             if eu.AWARD_PROP_NAME not in payload:
                 if not eu.AWARD:
                     raise AwardPropertyMissing
@@ -992,7 +1023,7 @@ class Connection():
         no_alias = False #Use this to check later if doing a GET
         aliases = payload.get(eu.ALIAS_PROP_NAME)
         if not aliases:
-            if profile_id in eup.Profile.NO_ALIAS_PROFILE_IDS or not require_aliases:
+            if not profile.has_alias or not require_aliases:
                 aliases = ["N/A"]
                 no_alias = True
             else:
@@ -1005,7 +1036,7 @@ class Connection():
         ### This doesn't work as locally I can't use jsonschema to validate a profile with
         ### custom objects specified in the value of a linkTo property.
         #self.debug_logger.debug("Validating the payload against the schema")
-        #validation_error = euu.err_context(payload=payload, schema=eup.Profile.PROFILES[profile_id])
+        #validation_error = euu.err_context(payload=payload, schema=self.profiles.get_profile_from_id(profile_id))
         #if validation_error:
         #    self.log_error("Invalid schema instance of the {} profile.".format(profile_id))
         #    self.debug_logger.debug("Payload is: {}".format(euu.print_format_dict(payload)))
@@ -1015,8 +1046,19 @@ class Connection():
         #    raise Exception(euu.print_format_dict(validation_error[0]))
 
         self.debug_logger.debug(
-            ("<<<<<< POST {} record {alias} To DCC with URL {url} and this"
-             " payload:\n\n{payload}\n\n").format(profile_id, alias=aliases[0], url=url, payload=euu.print_format_dict(payload)))
+            (
+                "<<<<<< POST {} record {alias} To DCC with URL {url} and this payload:"
+                "\n\n{payload}\n\n"
+            ).format(
+                profile.name,
+                alias=aliases[0],
+                url=url,
+                payload=euu.print_format_dict(
+                    payload,
+                    truncate_long_strings=truncate_long_strings_in_payload_log
+                )
+            )
+        )
 
         if self.check_dry_run():
             return {}
@@ -1028,6 +1070,7 @@ class Connection():
                                  verify=False)
         #response_json = response.json()["@graph"][0]
         response_json = response.json()
+        original_status_code = response.status_code
 
         if response.ok:
             self.debug_logger.debug("Success.")
@@ -1040,7 +1083,9 @@ class Connection():
                 encid = response_json["uuid"]
             self._log_post(aliases=aliases, dcc_id=encid)
             # Run 'after' hooks:
-            self.after_submit_hooks(encid, profile_id, method=self.POST)
+            self.after_submit_hooks(encid, profile.name, method=self.POST, upload_file=upload_file)
+            if return_original_status_code is True:
+                return (response_json, original_status_code)
             return response_json
         elif response.status_code == requests.codes.CONFLICT:
             self.debug_logger.debug(response_json)
@@ -1066,6 +1111,8 @@ class Connection():
                     response.raise_for_status()
                 else:
                     self.log_error("Will not POST '{}' since it already exists with aliases '{}'.".format(aliases[0], existing_record["aliases"]))
+                    if return_original_status_code is True:
+                        return (existing_record, original_status_code)
                     return existing_record
 
         else:
@@ -1123,11 +1170,18 @@ class Connection():
                     # not all props are pulled back when they are empty.
                     # For ex, in a file object, if the controlled_by prop isn't set, then
                     # it won't be in the response.
- 
+
                     ## CHECK FOR DUPLICATES: Be careful as some can be tricky, i.e.
-                    # ['/documents/id1', 'id1'] 
-                    # such a duplicate should be identified and removed, leaving us with ["id1"]. 
-                    payload[key] = eup.remove_duplicate_associations(val)
+                    # ['/documents/id1', 'id1']
+                    # such a duplicate should be identified and removed, leaving us with ["id1"].
+                    # Checks for arrays of strings or of dicts.
+                    if len(val) == 0:
+                        continue
+                    if isinstance(val[0], str):
+                        profile_id = self.get_profile_from_payload(payload)
+                        payload[key] = self.profiles.remove_duplicate_associations(val)
+                    elif isinstance(val[0], dict):
+                        payload[key] = euu.remove_duplicate_objects(val)
 
         # Run 'before' hooks:
         payload = self.before_submit_hooks(payload, method=self.PATCH)
@@ -1152,7 +1206,7 @@ class Connection():
             self.debug_logger.debug("Success.")
             response_json = response_json["@graph"][0]
             uuid = response_json["uuid"]
-            profile_id = eup.Profile(response_json["@id"]).profile_id
+            profile_id = self.profiles.get_profile_from_id(response_json["@id"]).name
             # Run 'after' hooks:
             self.after_submit_hooks(uuid, profile_id, method=self.PATCH)
             return response_json
@@ -1186,7 +1240,7 @@ class Connection():
         """
         self.debug_logger.debug("\nIN remove_props()")
         rec_json = self.get(rec_ids=rec_id, frame="object", ignore404=False)
-        profile = eup.Profile(rec_json["@id"])
+        profile = self.profiles.get_profile_from_id(rec_json["@id"])
         del rec_json
         editable_json = self.get(rec_ids=rec_id, frame="edit", ignore404=False)
         # For good house-keeping, check for any props that we definitely aren't allowed to remove,
@@ -1195,12 +1249,13 @@ class Connection():
         # the Portal will most likely complain or silently disallow an attempt to remove such
         # properites. Nonetheless, a well-behaved client shouldn't send uncouth requests, so some
         # checking is performed below for good measure:
-        for prop in props:
-            if profile.is_prop_required(prop):
+        for prop_name in props:
+            prop = profile.get_property_from_name(prop_name)
+            if prop.is_required:
                 raise Exception("Can't remove required property")
-            elif profile.is_prop_not_submittable(prop):
+            elif prop.is_not_submittable:
                 raise Exception("Can't remove non-submittable property.")
-            elif profile.is_prop_read_only(prop):
+            elif prop.is_read_only:
                 raise Exception("Can't remove read-only property.")
             else:
                 # Then it is safe to remove this property.
@@ -1223,6 +1278,146 @@ class Connection():
         response_json = response.json()
         return response_json
 
+    def remove_and_patch(
+        self,
+        props,
+        patch,
+        raise_403=True,
+        extend_array_values=True
+    ):
+        """Runs a PUT request to remove properties and patch a record in one
+        request.
+
+        In general, this is a method combining ``remove_props`` and ``patch``.
+        This is useful because some schema dependencies requires property
+        removal and property patch (including adding new properties) happening
+        at the same time. Please note that after the record is retrieved from
+        the portal, ``props`` will be removed before the ``patch`` is applied.
+
+        Args:
+            props: `list`. The properties to remove from the record.
+            patch: `dict`. containing the attribute key and value pairs to
+                patch. Must contain the key ``self.ENCID_KEY`` in order to
+                indicate which record to PATCH.
+            raise_403: `bool`. `True` means to raise a
+                ``requests.exceptions.HTTPError`` if a 403 status (forbidden)
+                is returned. If set to `False` and there still is a 403 return
+                status, then the object you were trying to PATCH will be
+                fetched from the Portal in JSON format as this function's
+                return value.
+            extend_array_values: `bool`. Only affects keys with array values.
+                `True` (default) means to extend the corresponding value on the
+                Portal with what's specified in the payload. `False` means to
+                replace the value on the Portal with what's in the payload.
+
+        Returns:
+            `dict`: The JSON response from the PUT operation, or an empty dict
+                if the record doesn't exist on the Portal. Will also be an
+                empty dict if the dry-run feature is enabled.
+
+        Raises:
+            requests.exceptions.HTTPError: if the return status is not ok
+                (excluding a 403 status if 'raise_403' is False).
+        """
+        if not props:
+            raise ValueError(
+                'Input props to remove must be non-empty. If you only need to '
+                'patch and not remove properties, use the patch() method of '
+                'the Connection class.'
+            )
+        if set(patch.keys()) <= {self.ENCID_KEY, self.PROFILE_KEY}:
+            raise ValueError(
+                'Input patch has no valide properties to patch. If you only '
+                'need to patch and not remove properties, use the '
+                'remove_props() method of the Connection class.'
+            )
+        # Make sure we have a payload that can be converted to valid JSON, and
+        # tuples become arrays, ...
+        patch = json.loads(json.dumps(patch))
+        self.debug_logger.debug("\nIN remove_and_patch()")
+        encode_id = patch[self.ENCID_KEY]
+        rec_json = self.get(rec_ids=encode_id, frame="object", ignore404=True)
+        if not rec_json:  # Ensure that the record exists on the Portal:
+            return {}
+        profile = self.profiles.get_profile_from_id(rec_json["@id"])
+        payload = self.get(rec_ids=encode_id, frame="edit", ignore404=False)
+        for prop_name in props:
+            prop = profile.get_property_from_name(prop_name)
+            if prop.is_required:
+                raise Exception("Can't remove required property")
+            elif prop.is_not_submittable:
+                raise Exception("Can't remove non-submittable property.")
+            elif prop.is_read_only:
+                raise Exception("Can't remove read-only property.")
+            else:
+                # Then it is safe to remove this property.
+                payload.pop(prop, None)
+        for key in patch:
+            if all([
+                extend_array_values,
+                key in payload,
+                isinstance(patch[key], list),
+            ]):
+                val = payload[key]
+                val.extend(patch[key])
+                # CHECK FOR DUPLICATES: Be careful as some can be tricky, i.e.
+                # ['/documents/id1', 'id1'] such a duplicate should be
+                # identified and removed, leaving us with ["id1"].
+                # Checks for arrays of strings or of dicts.
+                if len(val) == 0:
+                    continue
+                if isinstance(val[0], str):
+                    payload[key] = profile.remove_duplicate_associations(val)
+                elif isinstance(val[0], dict):
+                    payload[key] = euu.remove_duplicate_objects(val)
+            else:
+                payload[key] = patch[key]
+
+        # Run 'before' hooks:
+        payload = self.before_submit_hooks(payload, method=self.PATCH)
+        payload.pop(self.ENCID_KEY)
+        payload.pop(self.PROFILE_KEY, None)
+
+        url = euu.url_join([self.dcc_url, encode_id.lstrip("/")])
+        self.debug_logger.debug(
+            "<<<<<< PUTTING {encode_id} To DCC with URL"
+            " {url} and this payload:\n\n{payload}\n\n".format(
+                encode_id=encode_id,
+                url=url,
+                payload=euu.print_format_dict(payload)
+            )
+        )
+
+        if self.check_dry_run():
+            return {}
+        response = requests.put(
+            url,
+            auth=self.auth,
+            timeout=eu.TIMEOUT,
+            headers=euu.REQUEST_HEADERS_JSON,
+            json=payload,
+            verify=False
+        )
+        response_json = response.json()
+
+        if response.ok:
+            self.debug_logger.debug("Success.")
+            response_json = response_json["@graph"][0]
+            uuid = response_json["uuid"]
+            profile_id = self.profiles.get_profile_from_id(response_json["@id"]).name
+            # Run 'after' hooks:
+            self.after_submit_hooks(uuid, profile_id, method=self.PATCH)
+            return response_json
+        elif response.status_code == requests.codes.FORBIDDEN:
+            # Don't have permission to PATCH this object.
+            if not raise_403:
+                return rec_json
+
+        message = "Failed to PUT {}".format(encode_id)
+        self.log_error(message)
+        self.debug_logger.debug("<<<<<< DCC PUT RESPONSE: ")
+        self.debug_logger.debug(euu.print_format_dict(response_json))
+        response.raise_for_status()
 
     def send(self, payload, error_if_not_found=False, extend_array_values=True, raise_403=True):
         """
@@ -1585,18 +1780,18 @@ class Connection():
         file_rec = self.get(rec_ids=file_id,ignore404=False)
         if not file_path:
             try:
-                file_path = file_rec[eup.Profile.SUBMITTED_FILE_PROP_NAME]
+                file_path = file_rec[self.profiles.SUBMITTED_FILE_PROP_NAME]
             except KeyError:  # submitted_file_name property not set:
                 raise Exception("No file path specified.")
-        file_rec_md5sum = file_rec.get(eup.Profile.MD5SUM_NAME_PROP_NAME)
+        file_rec_md5sum = file_rec.get(self.profiles.MD5SUM_NAME_PROP_NAME)
         if not file_rec_md5sum or set_md5sum:
             # md5sum calc. supported at present only for local files and aws (not GCP)
             self.debug_logger.debug("Calculating md5sum for {}".format(os.path.basename(file_path)))
             md5sum = euu.calculate_md5sum(file_path)
             file_size = euu.calculate_file_size(file_path)
             self.patch({self.ENCID_KEY: file_rec["@id"], 
-                        eup.Profile.MD5SUM_NAME_PROP_NAME: md5sum,
-                        eup.FILE_SIZE_PROP_NAME: file_size})
+                        self.profiles.MD5SUM_NAME_PROP_NAME: md5sum,
+                        self.profiles.FILE_SIZE_PROP_NAME: file_size})
 
         cmd_args = "{file_path} {upload_url}".format(file_path=file_path, upload_url=aws_creds["UPLOAD_URL"])
         if file_path.startswith("gs://"):
